# Adapted from the Berkeley Pacman projects
# ---------------
# Licensing Information:  You are free to use or extend these projects for
# educational purposes provided that (1) you do not distribute or publish
# solutions, (2) you retain this notice, and (3) you provide clear
# attribution to UC Berkeley, including a link to http://ai.berkeley.edu.
# 
# Attribution Information: The Pacman AI projects were developed at UC Berkeley.
# The core projects and autograders were primarily created by John DeNero
# (denero@cs.berkeley.edu) and Dan Klein (klein@cs.berkeley.edu).
# Student side autograding was added by Brad Miller, Nick Hay, and
# Pieter Abbeel (pabbeel@cs.berkeley.edu).

import sys
import timeit

import viz
import searchutil

####################
####################

# Unified implementation of GraphSearch
# Input: SearchProblem, frontier data structure
# Output: [path], # nodes pushed, set(nodes pushed)
def graphSearch(problem, frontier):
	# Your code here!
	searchutil.raiseNotDefined()

# Implements DFS via graphSearch
# !! Do not change !!
def dfs(problem):
	return graphSearch(problem, searchutil.Stack())

# Implements BFS via graphSearch
# !! Do not change !!
def bfs(problem):
	return graphSearch(problem, searchutil.Queue())

# Implements UCS via graphSearch
# !! Do not change !!
def ucs(problem):
	return graphSearch(problem, searchutil.PriorityQueueWithFunction(lambda node: node[2]))

# Implements A* via graphSearch
# !! Do not change !!
def astar(problem, heuristic):
	return graphSearch(problem, searchutil.PriorityQueueWithFunction(lambda node: node[2]+heuristic(node[1], problem)))

# Input: search function (e.g. dfs)
# Output: [path], # nodes pushed, search time in seconds, set(nodes pushed)
# !! Do not change !!
def searchWithStats(searchFn):
	tic = timeit.default_timer()
	path, nodecount, nodes = searchFn()
	toc = timeit.default_timer()
	
	return path, nodecount, toc-tic, nodes

####################
####################

# SearchProblem subclass to allow
# search algorithms to solve mazes
class MazeProblem(searchutil.SearchProblem):
	def __init__(self, maze, costFn, start, finish):
		self.maze = maze
		self.costFn = costFn
		self.start = start
		self.finish = finish
		# you can add more here
		# but do not change above
	
	# !! Do not change !!
	def getMaze():
		return self.maze
	
	# !! Do not change !!
	def getStartLocation(self):
		return self.start

	# !! Do not change !!
	def getFinishLocation(self):
		return self.finish
		
	def getStartState(self):
		# Your code here!
		searchutil.raiseNotDefined()

	def isGoalState(self, state):
		# Your code here!
		searchutil.raiseNotDefined()

	def getSuccessors(self, state):
		successors = []
		for action in [searchutil.Directions.NORTH, searchutil.Directions.SOUTH, searchutil.Directions.EAST, searchutil.Directions.WEST]:
			# Your code here!
			searchutil.raiseNotDefined()

		return successors

# Input: location, maze problem
# Output: taxi-cab/L1 distance from location
#         to maze finish location
# !! Do not change !!
def manhattanHeuristic(position, mazeProblem):
    xy1 = position
    xy2 = mazeProblem.getFinishLocation()
    return abs(xy1[0] - xy2[0]) + abs(xy1[1] - xy2[1])

####################
####################

# List of search method names
# and associated functions
# (useful for looping through stats)
# 
# You may wish to temporarily modify this
# while debugging/evaluating
METHODS = [
	("DFS", dfs),
	("BFS", bfs),
	("UCS", ucs),
	("A* w/ L1", lambda p:astar(p, manhattanHeuristic)),
]

# Method map, useful for method lookup
# !! Do not change !!
METHODS_MAP = {m[0]:m[1] for m in METHODS}

####################
####################

# Produces visualizations of the maze solved with a method
# 
# Input: [[maze row 1], [maze row 2],...], start location, finish location, 
#        search function, visualization multiplier, color map, visualization title
# Output: [path], MazeProblem
# 
# You may wish to temporarily modify this
# while debugging/evaluating
def solveMaze(maze, start, finish, searchFn, multiplier, colors, title='Solved Path'):
	problem = MazeProblem(maze, viz.colorCostFn(colors), start, finish)
	
	# Temporarily swap commenting on the following lines to debug MazeProblem + mazeAddPath
	path, nodecount, time, nodes = searchWithStats(lambda: searchFn(problem))
	# path, nodecount, time, nodes = ['East', 'South', 'South', 'East', 'East', 'East', 'East', 'North', 'North', 'North', 'West'], 0, 0, set()
	
	vizorig = viz.mazeMat(maze, multiplier, viz.colorMap(colors), start, finish)
	
	pathcost = viz.mazeAddPath(vizorig, multiplier, path, problem)
	viz.showAndWait(vizorig, "{} (cost={:,d}; nodes={:,d}; time={:.2f}s)".format(title, pathcost, nodecount, time))
	
	return path, problem

####################
####################

# Visually solves a supplied maze (using a variety of search methods)
# 
# Expects... python search.py <path to maze problem> <visualization multiplier> [method]
# 
# You may wish to temporarily modify this
# while debugging/evaluating
def main(argv):
	mazepath = argv[0]
	multiplier = int(argv[1])
	method = argv[2] if len(argv) > 2 and argv[2] in METHODS_MAP else None
	colors = viz.COST_KWRGB
	
	# parse the maze file
	start, finish, maze = searchutil.readMaze(mazepath)
	
	# visualize the maze, expanding pixels per the multiplier
	viz.showMaze(maze, start, finish, multiplier, colors)
	
	# loop through method(s)
	for algName, searchFn in METHODS if method is None else ((method, METHODS_MAP[method]),):
		solveMaze(maze, start, finish, searchFn, multiplier, colors, algName)
		
	# end visualization
	viz.close()

# Runs from the command line
if __name__ == "__main__":
	main(sys.argv[1:])
